function [cost,grad,features] = sparseAutoencoderCost(theta, visibleSize, hiddenSize, ...
                                                      lambda, sparsityParam, beta, data)

global AEOptions;
                                         
% visibleSize: the number of input units (probably 64) 
% hiddenSize: the number of hidden units (probably 25) 
% lambda: weight decay parameter
% sparsityParam: The desired average activation for the hidden units (denoted in the lecture
%                           notes by the greek alphabet rho, which looks like a lower-case "p").
% beta: weight of sparsity penalty term
% data: Our 64x10000 matrix containing the training data.  So, data(:,i) is the i-th training example. 
  
% The input theta is a vector (because minFunc expects the parameters to be a vector). 
% We first convert theta to the (W1, W2, b1, b2) matrix/vector format, so that this 
% follows the notation convention of the lecture notes. 

W1 = reshape(theta(1:hiddenSize*visibleSize), hiddenSize, visibleSize);
W2 = reshape(theta(hiddenSize*visibleSize+1:2*hiddenSize*visibleSize), visibleSize, hiddenSize);
b1 = theta(2*hiddenSize*visibleSize+1:2*hiddenSize*visibleSize+hiddenSize);
b2 = theta(2*hiddenSize*visibleSize+hiddenSize+1:end);

% Cost and gradient variables (your code needs to compute these values). 
% Here, we initialize them to zeros. 
cost = 0;
W1grad = zeros(size(W1)); 
W2grad = zeros(size(W2));
b1grad = zeros(size(b1)); 
b2grad = zeros(size(b2));

%% ---------- YOUR CODE HERE --------------------------------------
%  Instructions: Compute the cost/optimization objective J_sparse(W,b) for the Sparse Autoencoder,
%                and the corresponding gradients W1grad, W2grad, b1grad, b2grad.
%
% W1grad, W2grad, b1grad and b2grad should be computed using backpropagation.
% Note that W1grad has the same dimensions as W1, b1grad has the same dimensions
% as b1, etc.  Your code should set W1grad to be the partial derivative of J_sparse(W,b) with
% respect to W1.  I.e., W1grad(i,j) should be the partial derivative of J_sparse(W,b) 
% with respect to the input parameter W1(i,j).  Thus, W1grad should be equal to the term 
% [(1/m) \Delta W^{(1)} + \lambda W^{(1)}] in the last block of pseudo-code in Section 2.2 
% of the lecture notes (and similarly for W2grad, b1grad, b2grad).
% 
% Stated differently, if we were using batch gradient descent to optimize the parameters,
% the gradient descent update to W1 would be W1 := W1 - alpha * W1grad, and similarly for W2, b1, b2. 
% 

% ------------ Sample Solution ------------

M = size(data, 2);

% Forward Prop
h = sigmoid(bsxfun(@plus, W1*data, b1));
features = h;

if AEOptions.linearAutoencoder
    r = bsxfun(@plus, W2*h, b2);
else
    r = sigmoid(bsxfun(@plus, W2*h, b2));
end

% Sparsity Cost
p = sparsityParam;
q = mean(h, 2);
cost = cost + sum(beta * ( p*log(p./q) + (1-p)*log((1-p)./(1-q))));

% Reconstruction Loss and Back Prop
diff = r - data;
cost = cost + 1/M * 0.5 * sum(diff(:).^2);
outderv = 1/M * diff;

% Output Layer
if AEOptions.linearAutoencoder
    outderv = outderv;
else
    outderv = outderv .* r .* (1 - r);
end

W2grad = outderv * h';
b2grad = sum(outderv, 2);
outderv = W2' * outderv ...
          + 1/M * repmat(beta * (-p./q + (1-p) ./ (1-q)), 1, M);

% Hidden Layer
outderv = outderv .* h .* (1 - h);
W1grad = outderv * data';
b1grad = sum(outderv, 2);

% Weight Regularization
cost = cost + sum(lambda * 0.5 * (W1(:).^2)) ...
            + sum(lambda * 0.5 * (W2(:).^2));
W1grad = W1grad + lambda * W1;
W2grad = W2grad + lambda * W2;
% ---------- End Sample Solution ---------- 



%-------------------------------------------------------------------
% After computing the cost and gradient, we will convert the gradients back
% to a vector format (suitable for minFunc).  Specifically, we will unroll
% your gradient matrices into a vector.

grad = [W1grad(:) ; W2grad(:) ; b1grad(:) ; b2grad(:)];

end
