function [cost,grad,activation,pred] = stackedAutoencoderCost(theta, sizeVector, classifierCost, ...
                                                         lambda, sparsityParam, beta, data, labels)

global AEOptions;

                                        
% sizeVector: vector of layer sizes; 
%             layer 1 = input layer; INCLUDES classifier layer
% lambda: weight decay parameter
% sparsityParam: The desired average activation for the hidden units (denoted in the lecture
%                           notes by the greek alphabet rho, which looks like a lower-case "p").
% beta: weight of sparsity penalty term
% data: Our 64x10000 matrix containing the training data.  So, data(:,i) is the i-th training example. 
  
% Extract parameters from theta

[numLayers, W, b] = stackedRollOutOfTheta(sizeVector, theta);

if isequal(classifierCost, 'sparseAE')
    %assert(numLayers == 3, 'Only 3 layers for sparseAE');
    assert(sizeVector(numLayers) == sizeVector(numLayers-2), 'Layer sizes must be appropriately set for sparseAE');
end

% Cost and gradient variables (your code needs to compute these values). 
% Here, we initialize them to zeros. 
cost = 0;
Wgrad = cell(numLayers, 1);
bgrad = cell(numLayers, 1);
for i = 1:numLayers-1
    Wgrad{i} = zeros(size(W{i}));
    bgrad{i} = zeros(size(b{i}));
end

%% ---------- YOUR CODE HERE --------------------------------------
%  Instructions: Compute the cost/optimization objective J_sparse(W,b) for the Sparse Autoencoder,
%                and the corresponding gradients W1grad, W2grad, b1grad, b2grad.

% ------------ Sample Solution ------------

M = size(data, 2);
outputMatrix = full(sparse(labels, 1:M, 1));

activation = cell(numLayers, 1);
incoming = cell(numLayers, 1);

% =============================
%          Forward prop
% =============================
%
% remember, numLayers INCLUDES the classifier layer!
%
% layer 1 = INPUT layer,
% activation{1} = activation for INPUT layer (which is just the data)
% incoming{1} = incoming for INPUT layer (which is also just the data)
%
% activation{k} = activation for the kth layer
% incoming{k} = incoming for the kth layer
%
h = data;
activation{1} = h;
incoming{1} = h;
for i = 2:numLayers 
    h = bsxfun(@plus, W{i-1} * h, b{i-1});
    incoming{i} = h;
    h = sigmoid(h);
    activation{i} = h;
end

% Note: incoming, NOT activation. The classifier works on the raw inputs,
% not on the raw inputs passed through sigmoid.
if isequal(classifierCost, 'sparseAE') 
    % then this is a "sparseAE layer" and we want its activation to be the
    % same as the activation of the layer before the hidden layer
    %{
      o o o o o <- this layer: classifier / "sparseAE layer"
      \ | | | /
        o o o   <- previous layer: "hidden layer"
      / | | | \
      o o o o o <- before layer: "input layer" 
    
      we want ACTIVATION of "input layer" to be equal to
              ACTIVATION of "sparseAE layer"
    %}            
    
    diff = activation{numLayers} - activation{numLayers-2};
    loss = (1/2) * sum(diff(:).^2);
    deriv = diff .* activation{numLayers} .* (1 - activation{numLayers});
            
    loss = loss / M;
    deriv = deriv / M;        
        
    % Sparsity Cost
    p = sparsityParam;
    q = mean(activation{2}, 2);
    sparsityCost = sum(beta * ( p*log(p./q) + (1-p)*log((1-p)./(1-q))));        

    cost = cost + sparsityCost;                    
else
    [ell, deriv, pred] = classifierCost(incoming{numLayers}, W{numLayers-1}, outputMatrix);
    
    % Maximise the log-likelihood = minimise the negative of the log-likelihood
    loss = -ell;
    deriv = -deriv;
    
    % disp(loss);
end

% loss and deriv should have now been determined by the above block
cost = cost + loss;
delta = deriv;

% =============================
%          Backprop
% =============================

% Output (i.e. classifier) layer

Wgrad{numLayers-1} = delta * activation{numLayers-1}';
bgrad{numLayers-1} = sum(delta, 2);

for i = numLayers-1:-1:2
    delta = W{i}' * delta; % "suck" delta backwards
    if ( (isequal(classifierCost, 'sparseAE')) && (i == 2) )       
        % Sparsity penalty for hidden layer
        p = sparsityParam;
        q = mean(activation{2}, 2);        
        delta = delta + 1/M * repmat(beta * (-p./q + (1-p) ./ (1-q)), 1, M);
    end
    delta = delta .* activation{i} .* (1 - activation{i});
    Wgrad{i-1} = delta * activation{i-1}';
    bgrad{i-1} = sum(delta, 2);
end
      
% Weight Regularization
for i = 1:numLayers-1    
    cost = cost + (lambda / 2) * sum(W{i}(:).^2);  
    Wgrad{i} = Wgrad{i} + lambda * W{i};
end
% ---------- End Sample Solution ----------

%-------------------------------------------------------------------
% After computing the cost and gradient, we will convert the gradients back
% to a vector format (suitable for minFunc).  Specifically, we will unroll
% your gradient matrices into a vector.

if isequal(classifierCost, @visualisationClassifierCost)
    % To visualise, we want the derivatives with respect to x, the input
    delta = W{1}' * delta;    
    epsilon = 1e-5;
    grad = delta + 1e-3 * data ./ sqrt(data(:).^2 + epsilon);    
    cost = loss + 1e-3 * sum(sqrt(data(:).^2 + epsilon));
else    
    grad = stackedRollIntoTheta(sizeVector, Wgrad, bgrad);
end    

end

